#script to process contents of mods folder in order to automatically upgrade mods
#broken by new release
#makes new folder new_mods
#reads .lua and .json and uses regexp to replace obsolete code
from os import walk
from os.path import isdir
from os import mkdir
from shutil import copy
from string import ascii_letters as ascii
import re

migration='0.12.35-0.13.0'
#creation, copying and writing of files
def main():#will be called at the end of file
    if isdir("./new_mods"):
        print('You should make "new_mods" directory disappear before running the script')
        exit()
    mkdir('./new_mods')
    
    print('updating')
    for (_,mods,_) in walk('mods'):
        break
    #here, we get major steps for lua updating:
    #in 12.35-13.0 for example:
    # 1. Simple renaming without regex syntax
    # 2. Renaming with regex, performs search in dictionary for every match
    # 3. Semi-manual code analysis, python regex can't express nested parenthesis 
    transformations =init_expressions_for(migration)
    modified=[]
    for m in mods:
        print('_____' + m + '_____')
        match=re.search(r"(.*(\d+?\.\d+?\.))(\d+?)$",m)
        if not match.group(2) or not match.group(3):
            print('modname format incorrect, skipping')
            continue
        #old version
        v=match.group(2)+match.group(3)
        #new folder name
        n=match.expand(match.group(1)+str(1+int(match.group(3))))
        #new version
        nv=match.expand(match.group(2)+str(1+int(match.group(3))))
        for (dirpath, dirnames, filenames) in walk('mods/'+m):
            path=dirpath.replace(m,n,1).replace('mods','new_mods',1)
            mkdir(path)
            for fn in filenames:
                #whether update did some replacements
                significant=False
                print('doing ',fn)
                match=re.search(r"\.(lua)|(info\.json)",fn)
                if match:
                    f=open(dirpath + '/' + fn)
                    code=f.read()
                    f.close()
                    nf=open(path+'/'+fn,'w')
                    if  fn=='info.json':
                        new_code=code.replace(v,nv,1)
                    else:
                        new_code=code
                        for t in transformations:#apply all specified modifications
                            new_code=t(new_code)
                        if not significant and new_code!=code: significant=True#check if changes made
                    nf.write(new_code)
                    nf.close()
                else:
                    copy(dirpath+'/'+fn,path+'/')
                #end if match
                if significant: modified.append(path+'/'+fn)
            #end for fn in filenames:
        #end for (dirpath, dirnames, filenames) in walk('mods/'+m):
    #end for m in mods:
    if len(modified)>0:
        print('Following files have been modified')
        for k in modified:print(k)

#pattern definition, no system calls below this line
migrations={}
def init_expressions_for(n):
    if migrations[n]:
        return migrations[n]()
    else:
        print('Unknown migration specified in code')
        exit()
        
#usable funcs, used to generate transformations of file
#put their outputs in list of tuples and it'll be fed to script
def simple_renaming(simple):
    #initializator for easy processings
    #takes dictionary of exact replacements
    #returns tupple of regex obj and replace function 
    string1='(?aims)('
    for k in simple:
        string1=string1+re.escape(k)+'|'
    string1=string1+')'
    simpl=re.compile(string1)
    def srep(m):
        if m.group(1)!='': 
            print('\treplacing ', m.group(1), ' with ',simple[m.group(1)])
            return simple[m.group(1)]
    def transform(code):
        return re.sub(simpl,srep,code)
    return transform
    
def medium_replacement(medium):
    #medium difficulty processing
    #takes dictionary with regex syntax
    string2='(?aims)'
    i=1;
    mres={}
    for k,v in medium.items():
        string2=string2+'(?P<'+ascii[i]+'>'+k+')|'
        mres[ascii[i]]=v
        i=i+1;
    def mrep(m):
        for k,v in mres.items():
            if m.group(k): 
                print('\treplacing ', m.group(k), ' with ',v)
                return v 
    med=re.compile(string2)
    def transform(code):
        return re.sub(med,mrep,code)
    return transform

#handy for hardest replacements below
def count_lines(string, pos):
    #returns linenumber, where pos is located
    l=1
    lin=re.compile(r'\n')
    for m in lin.finditer(string,0,pos):l=l+1
    return l
def find_bra(string,pos=0,ope='(',clo=')'):
    #finds in 'string' the outermost brakets to the right of 'pos' specified by 'ope' and 'clo'
    #returns the bracket positions (left,right)
    reg=re.compile('(?P<left>\\'+ope+')|(?P<right>\\'+clo+')')
    num=0
    first=reg.search(string,pos)
    if first and first.group('left'): 
        first=first.start()
        num=num+1
    else: return None
    
    index=first+1
    while index < len(string) and num>0:
        m=reg.search(string,index)
        if m.group('left'):
            num=num+1
        elif m.group('right'):
            if num>1: num=num-1
            else: 
                return(first,m.end())
        index=m.end()
    return None
def find_out_bra(string,pos=0):
    #wrapper to find outermost of  any possible brakets
    bras={'(':')','[':']','{':'}'}
    res=()
    for k,v in bras.items():
        bp=find_bra(string,pos,k,v)
        if bp and( (not res) or res[0]>bp[0]): res=bp; 
    return res

#hardest replacements are here.
#they're specific function to perform a particular transformation over text
def ins_arg(hard):
    #returns func that
    #transforms the function calls () to table indexing [] and so on
    #takes a dictionary of dicts with particular fields (see #'0.12.35-0.13.0' example)
    bras={'(':')','[':']','{':'}'}
    string3='(?aims)'
    i=1;
    hres={}
    for k in hard:
        string3=string3+'(?P<'+ascii[i]+'>'+k['old']+'\W*?\\'+k['old_bra']+')|'
        hres[ascii[i]]=k
        i=i+1;
    
    reg=re.compile(string3)
    
    def transform(code):
        #in current modification is fit for single character brakets only
        index=0#reader position
        ms=[]#modifications to perform
        lett=list(code)#will do modifications on list of letters 
        while index<len(code):#scan code for names to replace
            m=reg.search(code,index)
            index=index+1
            if not m: break
            for k,v in hres.items():
                if m.group(k):
                    print('\tchanging parenthis',m.group(k))
                    index=m.end()-1
                    r=bras[v['old_bra']]
                    bp=find_bra(code,pos=index,ope=v['old_bra'],clo=r)#braket positions
                    if not bp:
                        print ('parenthises fail');
                        print(count_lines(code,index))
                        exit()#error
                    ope=v['new_bra']; clo=bras[ope];#brakets are substituted in place
                    lett[bp[0]]=ope
                    lett[bp[1]-1]=clo
                    ms.append((v,(m.start(),m.end())))#memorize what names to replace
        for i in range(len(ms)-1,-1,-1):
            v=ms[i]
            del(lett[v[1][0]:v[1][1]-1])
            lett.insert(v[1][0],v[0]['new'])
        new_code=''.join(lett)
        return new_code            
    return transform

def shuffle_args(hard):
    #returns function that finds function with two args and swaps places of args
    bras={'(':')','[':']','{':'}'}
    string4='(?aims)'
    i=1;
    hres={}
    for k in hard:
        string4=string4+'(?P<'+ascii[i]+'>'+k['old']+'\W*?\\'+k['old_bra']+')|'
        hres[ascii[i]]=k
        i=i+1;
    reg=re.compile(string4)
    def shuff(code):
        index=0#reader position
        modifications=[]#modifications to perform
        lett=list(code)#will do modifications on list of letters 
        comma=re.compile(',')
        while index<len(code):#scan code for names to replace
            m=reg.search(code,index)
            index=index+1
            if not m: break
            for k,v in hres.items():
                if m.group(k):
                    print(m.group(k))
                    print('\tshuffling arguments',m.group(k))
                    index=m.end()
                    l=v['old_bra']
                    r=bras[l]
                    bp=find_bra(code,pos=index-1,ope=l,clo=r)#braket positions
                    if not bp:
                        print ('parenthises fail on line: ',count_lines(code,index))
                        exit()#error
                    #count commas
                    commas=[];
                    for x in comma.finditer(code,bp[0]+1,bp[1]): commas.append(x.start());
                    err=False
                    if len(commas)==0: err=True
                    if len(commas)>0: #ms.append((bp[0]+1,p[0]-1,bp[2]));
                    #filter out commas in nested brakets
                        brakets=[]
                        ind=bp[0]+1
                        while ind<bp[1]:
                            bp2=find_out_bra(code,pos=ind)
                            if not bp2 or bp2[1]>bp[1]: break
                            commas=[x for x in commas if (x<bp2[0] or x>bp2[1]-1)]
                            if len(commas)==0: err=True; break
                            ind=bp2[1]
                        if len(commas)>1: err=True
                    if err: print('error, shuffle called on not two arg fun ',count_lines(code,index))#will just move on
                    else:
                        modifications.append((bp[0]+1,commas[0],bp[1]-1));
                        print('\tshuffling arguments in ',m.group(k))
                #end if m.group(k):
            #end for k,v in hres.items():
        #end while index<len(code)
        for m in reversed(modifications):
            left=code[m[0]:m[1]]
            right=code[m[1]+1:m[2]]
            del(lett[m[1]+1:m[2]])
            lett.insert(m[1]+1,left)
            del(lett[m[0]:m[1]])
            lett.insert(m[0],right)
        new_code=''.join(lett)
        return new_code
    #end def shuff
    return shuff
#end def shuffle_args(hard)
                            
    
#the actual pattern definition
def _():#'0.12.35-0.13.0'
    #list of changes
    simple={
        #these substituitions are simple and even if recognized wrongly, will not break most things
        #no regex in them
        'basic-transport-belt':'transport-belt',
        'transport-belt-to-ground':'underground-belt',
        'basic-splitter':'splitter',
        'basic-inserter':'inserter',
        'basic-mining-drill':'electric-mining-drill',
        'basic_beacon':'beacon', 
        'basic-bullet-magazine':'firearm-magazine', 
        'piercing-bullet-magazine':'piercing-rounds-magazine',
        'basic-grenade':'grenade',
        'basic-armor':'light-armor', 
        'basic-modular-armor':'modular-armor', 
        'basic-laser-defense-equipment':'personal-laser-defense-equipment',
        'basic-exoskeleton-equipment':'exoskeleton-equipment',
        'groupstate':'group_state',
        'circuitconnector':'circuit_connector', 
        'circuitconditionindex':'circuit_condition_index', 
        'trainstate':'train_state',
        'logistic_robot_storage_modifier':'logistic_robot_storage_bonus',
        }
    medium={
        #these are needing named groups to be replaced
        'require\W*?\'defines(\.lua)*\'':'',
        'require\W*?\"defines(\.lua)*\"':'',
        'game\.player(?!s)':'game.players[1]',
        'game\.local_player':'game.player',
        'game.regenerate_tiles\(\)':'',
        'ipairs\W*\(game.players':'pairs\W*\(\Wgame.players',
        'ipairs\W*\(\W*game.surfaces':'pairs\(game.surfaces',
        'ipairs\W*\(\W*game.entity_prototypes':'pairs\(game.entity_prototypes',
        'ipairs\W*\(\W*game.item_prototypes':'pairs\(game.item_prototypes',
        'ipairs\W*\(\W*game.fluid_prototypes':'pairs\(game.fluid_prototypes',
        'ipairs\W*\(\W*force.recipes':'pairs\(force.recipes',
        'ipairs\W*\(\W*force.technologies':'pairs\(force,technologies',
        }
    hard=[
        #python cant find balanced parenteses will have to be done by hand
        #each of these has a potential to get personal replacer function
        {'old':'get_filter','new':'get_filter','transform':'shuffle','old_bra':'(','new_bra':'('},
        {'old':'set_filter','new':'set_filter','transform':'shuffle','old_bra':'(','new_bra':'('},
        {'old':'game.get_player','new':'game.players','transform':'insert_arg','old_bra':'(','new_bra':'['},
        {'old':'game.get_surface','new':'game.surfaces','transform':'insert_arg','old_bra':'(','new_bra':'['}
    ]
    actions=[]
    actions.append(simple_renaming(simple))
    actions.append(medium_replacement(medium))
    h1=[]; h2=[];
    for k in hard:
        if k['transform']=='insert_arg':
            h1.append(k)
        elif k['transform']=='shuffle':
            h2.append(k)
    actions.append(ins_arg(h1))
    #actions.append(shuffle_args(h2))
    
    return actions

migrations['0.12.35-0.13.0']=_

main()
